'''
App class.
'''

import inspect
from unstack.app.defaults import CONFIG, ROUTES
from unstack.app.logger   import Logger
from unstack.http.server  import Server
from unstack.util.daemon  import Daemon
from unstack.http.data.response import Response

class App(Daemon):
	'''
	A single web application.
	'''

	def __init__(self, config=None):
		'''
		Initialise a new App.
		'''

		self.headers = {}
		self.config  = config or CONFIG
		self.routes  = ROUTES
		self.server  = Server(self.config, self.serve)
		self.address = self.server.address
		self.config['logger'] = Logger(self.config)
		if self.config['daemonise']:
			super().__init__(self.config)
		self.config['logger']('App initialised.')
		self.config['logger']('App now listening on {}:{}.', *self.address)

	def __contains__(self, route):
		'''
		Return True if the route is in self.routes.
		'''

		return route in self.routes

	def route(self, verb, path, **headers):
		'''
		Decorate a function to add it as a route.
		'''

		def deco(func):
			self.routes[(verb, path)]  = func
			self.headers[(verb, path)] = headers
			return func
		return deco

	get  = lambda self, path, **hdrs: self.route('GET',  path, **hdrs)
	post = lambda self, path, **hdrs: self.route('POST', path, **hdrs)

	def error(self, code, **headers):
		'''
		Decorate a function to add it as an error handler.
		'''

		def deco(func):
			self.routes[code]  = func
			self.headers[code] = headers
			return func
		return deco

	def apply(self, route, request, error=None):
		'''
		Return the value of a route function applied to a Request, only
		passing arguments that the function possesses.
		'''

		func = self.routes[route]
		args = tuple(inspect.getfullargspec(func).args)
		return {
			():                   lambda: func(),
			('error', ):          lambda: func(error),
			('request', ):        lambda: func(request),
			('request', 'error'): lambda: func(request, error),
		}[args]()

	def serve(self, request):
		'''
		Return a routed Response from an incoming Request.
		'''

		try:
			try:
				if request.route in self:
					code, body = 200, self.apply(request.route, request)
				else:
					code, body = 404, self.apply(404, request)
			except Exception as error:
				code, body = 500, self.apply(500, request, error)
		except Exception as error:
			code, body = 500, 'Error page suffered error: {}.'.format(error)

		headers = self.headers.get(request.route, {})
		return Response(request, code, body, headers)

	def run(self):
		'''
		Start and run the App's Server.
		'''

		try:
			if self.config['daemonise']:
				self.start_daemon()
			self.server.start()
		except KeyboardInterrupt:
			if self.config['daemonise']:
				self.stop_daemon()
			self.server.stop()
