'''
Headers class.
'''

class Headers:
	'''
	A dictionary of HTTP headers.
	'''

	def __init__(self, *datas):
		'''
		Initialise a new Headers object.
		'''

		self.headers = {}

		for data in datas:
			self.update(data)

	def __contains__(self, key):
		'''
		Return True if a key is in self.headers.
		'''

		return self.clean_key(key) in self.headers

	def __getitem__(self, key):
		'''
		Return a header value from self.headers.
		'''

		return self.headers[self.clean_key(key)]

	def __delitem__(self, key):
		'''
		Delete a header value from self.headers.
		'''

		del self.headers[self.clean_key(key)]

	def __setitem__(self, key, value):
		'''
		Set a header value into self.headers.
		'''

		self.headers[self.clean_key(key)] = value

	def __str__(self):
		'''
		Return the Headers data as a string.
		'''

		items = sorted(self.headers.items(), key=lambda pair: pair[0])
		return '\n'.join('{}: {}'.format(key, val) for key, val in items)

	def update(self, data):
		'''
		Update self.headers with new values from a string, dictionary
		or Header instance.
		'''

		if isinstance(data, str):
			data = self.parse_string(data)
		elif isinstance(data, (list, tuple)):
			data = self.parse_string('\n'.join(data))
		elif isinstance(data, Headers):
			data = data.headers

		self.headers.update(self.clean_dict(data))

	def clean_key(self, key):
		'''
		Return a header key in Header-Case.
		'''

		key = key.strip().replace('_', '-').replace(' ', '-')
		return '-'.join(word.title() for word in key.split('-'))

	def clean_dict(self, dct):
		'''
		Return a dictionary with keys in Header-Case and values stripped.
		'''

		return {self.clean_key(key):val.strip() for key, val in dct.items()}

	def parse_string(self, string):
		'''
		Return a dictionary from a string of headers.
		'''

		return self.clean_dict(
			dict(line.split(':', 1) for line in string.splitlines())
		)
