'''
Server class.
'''

import time
from unstack.http.socket import Socket

class Server:
	'''
	A function-handling socket server.
	'''

	def __init__(self, config, func):
		'''
		Initialise a new Server object.
		'''

		self.config  = config
		self.func    = func
		self.running = True
		self.connections = {}
		self.requests    = {}
		self.responses   = {}
		self.sock = Socket(config=self.config)
		self.address = self.sock.address

	def clear(self, address):
		'''
		Delete a completed connection from all registers.
		'''

		del self.connections[address]
		del self.requests[address]
		del self.responses[address]

	def sleep_if_empty(self):
		'''
		Sleep to wait for more connections if necessary.
		'''

		if not self.connections:
			time.sleep(self.config['wait_time'])

	def bind(self):
		'''
		Bind the server's Socket to begin listening for connections.
		'''

		self.sock.bind()

	def accept(self):
		'''
		Accept all incoming connections.
		'''

		while self.sock.can('recv'):
			conn = self.sock.accept()
			if conn:
				self.connections[conn.address] = conn
		self.sleep_if_empty()

	def request(self):
		'''
		Receive Requests from all existing connections.
		'''

		for address, conn in self.connections.items():
			request = conn.recv_request(self.config['recv_size'])
			if request:
				self.requests[address] = request

	def process(self):
		'''
		Process Responses from all existing Requests.
		'''

		for address, request in self.requests.items():
			self.responses[address] = self.func(request)

	def respond(self):
		'''
		Send Responses to all existing connections.
		'''

		done = []
		for address, response in self.responses.items():
			conn = self.connections[address]
			if conn.send_response(response):
				done.append(address)

		for address in done:
			self.config['logger']('{} => {}', conn, response.request)
			self.clear(address)

	def start(self):
		'''
		Start and run the Server.
		'''

		self.bind()

		while self.running:
			try:
				for func in ('accept', 'request', 'process', 'respond'):
					getattr(self, func)()
			except Exception as error:
				self.stop()
				raise error

	def stop(self):
		self.running = False
		self.sock.close()
		for conn in self.connections.values():
			conn.close()
		for data in ('connections', 'requests', 'responses'):
			getattr(self, data).clear()
