'''
Socket class.
'''

import select
import socket
from unstack.http.data.request import Request

def if_can(mode):
	'''
	Decorate a Socket method to run only if self.can(mode) is True.
	'''

	def deco(func):
		def wrap(self, *args, **kwargs):
			return func(self, *args, **kwargs) if self.can(mode) else False
		return wrap
	return deco


class Socket:
	'''
	A socket connection.
	'''

	socket.setdefaulttimeout(0)

	def __init__(self, **kwargs):
		'''
		Initialise a new Socket object.
		'''

		if 'sock' in kwargs and 'address' in kwargs:
			self.sock = kwargs['sock']
			self.host, self.port = kwargs['address']

		elif 'host' in kwargs and 'port' in kwargs:
			self.sock = socket.socket()
			self.host = str(kwargs['host']) or '127.0.0.1'
			self.port = int(kwargs['port'])

		elif 'config' in kwargs:
			self.sock = socket.socket()
			self.host = str(kwargs['config']['host'])
			self.port = int(kwargs['config']['port'])

		self.address = self.host, self.port
		self.sock.setblocking(False)

	def __str__(self):
		'''
		Return the Socket as a string.
		'''

		return '{}:{}'.format(self.host, self.port)

	def can(self, mode):
		'''
		Return True if the socket is able to send/receive.
		'''

		if self.sock.fileno() != -1:
			sels = select.select([self.sock], [self.sock], [], 0)
			return self.sock in dict(zip(('recv', 'send'), sels))[mode]
		else:
			return False

	def bind(self):
		'''
		Bind the socket to a local address and listen for connections.
		'''

		self.sock.bind(self.address)
		self.sock.listen(1)

	def connect(self):
		'''
		Connect the socket to a remote address, with a blocking operating.
		'''

		self.sock.setblocking(True)
		self.sock.connect(self.address)
		self.sock.setblocking(False)

	@if_can('recv')
	def accept(self):
		'''
		Return a new Socket connection and return True, or False if cannot.
		'''

		conn, address = self.sock.accept()
		if conn and address:
			conn.setblocking(False)
			return Socket(sock=conn, address=address)
		else:
			return None

	@if_can('recv')
	def recv(self, size):
		'''
		Return a string from the socket, or False if cannot.
		'''

		data = self.sock.recv(size).decode().strip()
		return data or None

	@if_can('send')
	def send(self, data):
		'''
		Send a string into the socket and return True, or False if cannot.
		'''

		self.sock.send(data.encode())
		return True

	def close(self):
		'''
		Close the socket.
		'''

		self.sock.close()

	@if_can('recv')
	def recv_request(self, size):
		'''
		Return a Request object from the socket, or False if cannot.
		'''

		data = self.recv(size)
		return Request(data, self.address) if data else None

	@if_can('send')
	def send_response(self, response):
		'''
		Send a Response object into the socket and return True, or
		False if cannot.
		'''

		return self.send(str(response))
