'''
Daemon class.
'''

import atexit
import os
import pathlib
import signal
import sys
import time

class Daemon:
	'''
	A subclass-able daemon class.
	'''

	def __init__(self, config):
		'''
		Initialise a new Daemon object.
		'''

		self.config  = config
		self.pidfile = pathlib.Path(self.config['pid_path'])
		self.config['logger']('Daemon initialised.')
		atexit.register(self.delete_pid)

	def fork(self):
		'''
		Fork the daemon's process.
		'''

		try:
			pid = os.fork()
			if pid > 0:
				sys.exit(0)
		except OSError as error:
			self.config['logger']('Daemon fork failed: {}', error)
			sys.exit(1)

	def decouple(self):
		'''
		Decouple the daemon's process from the terminal.
		'''

		os.chdir('/')
		os.setsid()
		os.umask(0)

	def redirect(self):
		'''
		Redirect the daemon process's file descriptors.
		'''

		sys.stdout.flush()
		sys.stderr.flush()
		si = open(os.devnull, 'r')
		so = open(os.devnull, 'a+')
		se = open(os.devnull, 'a+')
		os.dup2(si.fileno(), sys.stdin.fileno())
		os.dup2(so.fileno(), sys.stdout.fileno())
		os.dup2(se.fileno(), sys.stderr.fileno())

	def read_pid(self):
		'''
		Read the current PID from the file if it exists.
		'''

		if self.pidfile.is_file():
			with self.pidfile.open('r') as fp:
				return int(fp.read().strip())
		else:
			return None

	def write_pid(self):
		'''
		Register the PID and write to the configured filepath.
		'''

		with self.pidfile.open('w') as fp:
			fp.write(str(os.getpid()) + '\n')

	def delete_pid(self):
		'''
		Delete the PID file if it exists.
		'''

		if self.pidfile.is_file():
			self.pidfile.unlink()

	def daemonize(self):
		'''
		Deamonize the object using the Unix double fork mechanism.
		'''


		for func in ('fork', 'decouple', 'fork', 'redirect', 'write_pid'):
			getattr(self, func)()
		self.config['logger']('App daemonised.')

	def start_daemon(self):
		'''
		Start the daemon.
		'''

		if sys.platform.startswith('win'):
			self.config['logger']('Cannot daemonise on Windows.')
		else:
			pid = self.read_pid()

			if pid:
				self.config['logger']('PID {} already exists. Exiting.', pid)
				sys.exit(1)
			else:
				self.daemonize()
				self.config['logger']('App daemon started.')
				self.run()

	def stop_daemon(self):
		'''
		Stop the daemon.
		'''

		pid = self.read_pid()

		if not pid:
			self.config['logger']('No existing PID file. Exiting.', pid)
		else:
			try:
				while True:
					os.kill(pid, signal.SIGTERM)
					time.sleep(0.1)
			except OSError as error:
				if self.pidfile.is_file():
					self.delete_pid()
				else:
					raise
			self.config['logger']('App daemon stopped.')

	def run(self):
		'''
		Subclass this method to provide daemon actions.
		'''

		pass
